#ifndef _HLL_PRIVATE_HPP_
#define _HLL_PRIVATE_HPP_

#include "AuxHashMap.hpp"
#include "CompositeInterpolationXTable.hpp"
#include "CouponHashSet.hpp"
#include "CouponList.hpp"
#include "CubicInterpolation.hpp"
#include "HarmonicNumbers.hpp"
#include "Hll4Array.hpp"
#include "Hll6Array.hpp"
#include "Hll8Array.hpp"
#include "HllArray.hpp"
#include "HllSketchImpl.hpp"
#include "HllSketchImplFactory.hpp"
#include "HllUtil.hpp"
#include "RelativeErrorTables.hpp"

#include "AuxHashMap-internal.hpp"
#include "coupon_iterator.hpp"
#include "CouponHashSet-internal.hpp"
#include "CouponList-internal.hpp"
#include "Hll4Array-internal.hpp"
#include "Hll6Array-internal.hpp"
#include "Hll8Array-internal.hpp"
#include "HllArray-internal.hpp"
#include "HllSketch-internal.hpp"
#include "HllSketchImpl-internal.hpp"
#include "HllUnion-internal.hpp"
#include "coupon_iterator-internal.hpp"

#endif // _HLL_PRIVATE_HPP_
